from collections.abc import Callable, Iterable

import torch
import torch.nn as nn
import torch.nn.functional as F

from .flow import Distribution
from .masked_mlp import MaskedMLP


class DiscreteMADE(nn.Module, Distribution):
    def __init__(
        self,
        dims_in: list[int],
        dims_c: int = 0,
        channels: int | None = None,
        prior_prob_function: Callable[[torch.Tensor, int], torch.Tensor] | None = None,
        channel_remap_function: Callable[[torch.Tensor], torch.Tensor] | None = None,
        **mlp_kwargs,  # TODO: replace this with default arguments for MLP
    ):
        """ """
        super().__init__()

        self.autoreg_net = MaskedMLP(
            input_dims=[dims_c, *dims_in[:-1]],
            output_dims=dims_in,
            channels=channels,
            **mlp_kwargs,
        )
        self.dims_in = dims_in
        self.max_dim = max(dims_in)
        self.channels = 1 if channels is None else channels
        self.prior_prob_function = prior_prob_function
        self.channel_remap_function = channel_remap_function

        discrete_indices = []
        one_hot_mask = []
        for i, dim in enumerate(dims_in):
            discrete_indices.extend([i] * dim)
            one_hot_mask.extend([True] * dim + [False] * (self.max_dim - dim))
        self.register_buffer("discrete_indices", torch.tensor(discrete_indices))
        self.register_buffer("one_hot_mask", torch.tensor(one_hot_mask))
        self.register_buffer("dims_in_tensor", torch.tensor(dims_in))
        self.register_buffer("dummy", torch.tensor([0.0]))

    def _sort_channels(
        self,
        channel: torch.Tensor | list[int] | int | None,
        tensors: Iterable[torch.Tensor | None, ...],
    ) -> tuple[
        list[int] | int | None, torch.Tensor | None, tuple[torch.Tensor | None, ...]
    ]:
        if isinstance(channel, torch.Tensor):
            if self.channel_remap_function is not None:
                channel = self.channel_remap_function(channel)
            channel_perm = torch.argsort(channel)
            channel_sizes = channel.bincount(minlength=self.channels).tolist()
            perm_tensors = tuple(
                tensor[channel_perm] if tensor is not None else None
                for tensor in tensors
            )
            return channel_sizes, channel_perm, perm_tensors

        if self.channel_remap_function is not None:
            if isinstance(channel, int):
                channel = self.channel_remap_function(channel)
            elif isinstance(channel, list):
                raise ValueError(
                    "channel_remap_function not supported if called "
                    "with list of channel sizes"
                )
        return channel, None, tensors

    def _unsort_channels(
        self,
        channel_perm: torch.Tensor | None,
        tensors: Iterable[torch.Tensor | None, ...],
    ) -> tuple[torch.Tensor | None, ...]:
        if channel_perm is None:
            return tensors
        channel_perm_inv = torch.argsort(channel_perm)
        return tuple(
            tensor[channel_perm_inv] if tensor is not None else None
            for tensor in tensors
        )

    def log_prob(
        self,
        x: torch.Tensor,
        c: torch.Tensor | None = None,
        channel: torch.Tensor | list[int] | int | None = None,
    ) -> torch.Tensor:
        """ """
        if self.prior_prob_function is not None:
            prior_probs = torch.cat(
                [
                    self.prior_prob_function(x[:, :i], i)
                    for i, _ in enumerate(self.dims_in)
                ],
                dim=1,
            )
        else:
            prior_probs = None

        x_one_hot = (
            F.one_hot(x, self.max_dim)
            .to(self.dummy.dtype)
            .flatten(start_dim=1)[:, self.one_hot_mask]
        )

        channel, channel_perm, (x_one_hot, c, prior_probs) = self._sort_channels(
            channel, (x_one_hot, c, prior_probs)
        )

        net_input = x_one_hot if c is None else torch.cat((c, x_one_hot), dim=1)
        net_prob = self.autoreg_net(net_input[:, : -self.dims_in[-1]], channel).exp()
        unnorm_prob = (
            net_prob if self.prior_prob_function is None else net_prob * prior_probs
        )

        prob_norms = torch.zeros_like(x, dtype=x_one_hot.dtype).scatter_add_(
            dim=1,
            index=self.discrete_indices[None, :].expand(x.shape[0], -1),
            src=unnorm_prob,
        )
        prob_sums = torch.zeros_like(prob_norms).scatter_add_(
            dim=1,
            index=self.discrete_indices[None, :].expand(x.shape[0], -1),
            src=unnorm_prob * x_one_hot,
        )
        prob = torch.prod(prob_sums / prob_norms, dim=1)

        return self._unsort_channels(channel_perm, (prob.log(),))[0]

    def sample(
        self,
        n: int | None = None,
        c: torch.Tensor | None = None,
        channel: torch.Tensor | list[int] | int | None = None,
        return_log_prob: bool = False,
        return_prob: bool = False,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ) -> torch.Tensor | tuple[torch.Tensor, ...]:
        """ """
        if n is None:
            n = len(c)
        if c is None:
            options = {"device": device, "dtype": dtype}
            x_in = torch.zeros((n, 0), **options)
        else:
            options = {"device": c.device, "dtype": c.dtype}
            x_in = c

        prob = torch.ones((n,), **options)
        net_cache = None

        channel, channel_perm, (x_in,) = self._sort_channels(channel, (x_in,))

        x_out = torch.zeros(
            (n, len(self.dims_in)),
            device=options["device"],
            dtype=torch.int64,
        )
        for i, dim in enumerate(self.dims_in):
            y, net_cache = self.autoreg_net.forward_cached(x_in, i, net_cache, channel)
            net_probs = y.exp()
            if self.prior_prob_function is None:
                unnorm_probs = net_probs
            else:
                prior_probs = self.prior_prob_function(x_out[:, :i], i)
                unnorm_probs = net_probs * prior_probs
            cdf = unnorm_probs.cumsum(dim=1)
            norm = cdf[:, -1]
            cdf = cdf / norm[:, None]
            r = torch.rand((y.shape[0], 1), **options)
            samples = torch.clip(torch.searchsorted(cdf, r)[:, 0], max=cdf.shape[1] - 1)
            x_out[:, i] = samples
            prob = prob * torch.gather(unnorm_probs, 1, samples[:, None])[:, 0] / norm
            x_in = F.one_hot(samples, dim).to(y.dtype)

        return_list = [x_out]
        if return_log_prob:
            return_list.append(prob.log())
        if return_prob:
            return_list.append(prob)
        return_list = self._unsort_channels(channel_perm, return_list)
        if len(return_list) > 1:
            return return_list
        else:
            return return_list[0]



from typing import Any, Literal

import torch
import torch.nn as nn
import torch.nn.functional as F

from .discrete_made import DiscreteMADE
from .discrete_transformer import DiscreteTransformer
from .flow import Distribution, Flow


class MixedFlow(nn.Module, Distribution):
    def __init__(
        self,
        dims_in_continuous: int,
        dims_in_discrete: list[int],
        dims_c: int = 0,
        discrete_dims_position: Literal["first", "last"] = "first",
        discrete_model: Literal["made", "transformer"] = "made",
        channels: int | None = None,
        continuous_kwargs: dict[str, Any] = {},
        discrete_kwargs: dict[str, Any] = {},
    ):
        """ """
        super().__init__()
        if discrete_dims_position == "first":
            self.discrete_dims_first = True
        elif discrete_dims_position == "last":
            self.discrete_dims_first = False
        else:
            raise ValueError("discrete_dims_position must be 'first' or 'last'")
        self.dims_in_continuous = dims_in_discrete
        self.dims_in_discrete = len(dims_in_discrete)
        if self.discrete_dims_first:
            dims_c_discrete = dims_c
            dims_c_continuous = sum(dims_in_discrete) + dims_c
        else:
            dims_c_discrete = dims_c + dims_in_continuous
            dims_c_continuous = dims_c

        if discrete_model == "made":
            self.discrete_flow = DiscreteMADE(
                dims_in=dims_in_discrete,
                dims_c=dims_c_discrete,
                channels=channels,
                **discrete_kwargs,
            )
        elif discrete_model == "transformer":
            self.discrete_flow = DiscreteTransformer(
                dims_in=dims_in_discrete,
                dims_c=dims_c_discrete,
                **discrete_kwargs,
            )
            if channels is not None:
                raise NotImplementedError(
                    "DiscreteTransformer only supported for single-channel integration"
                )
        else:
            raise ValueError("discrete_model must be 'made' or 'transformer'")

        self.continuous_flow = Flow(
            dims_in=dims_in_continuous,
            dims_c=dims_c_continuous,
            channels=channels,
            **continuous_kwargs,
        )

        self.max_dim_discrete = max(dims_in_discrete)
        one_hot_mask = []
        for dim in dims_in_discrete:
            one_hot_mask.extend([True] * dim + [False] * (self.max_dim_discrete - dim))
        self.register_buffer("one_hot_mask", torch.tensor(one_hot_mask))
        self.register_buffer("dummy", torch.tensor([0.0]))

    def encode_discrete(self, indices: torch.Tensor) -> torch.Tensor:
        return (
            F.one_hot(indices, self.max_dim_discrete)
            .to(self.dummy.dtype)
            .flatten(start_dim=1)[:, self.one_hot_mask]
        )

    def log_prob(
        self,
        x: torch.Tensor,
        c: torch.Tensor | None = None,
        channel: torch.Tensor | list[int] | int | None = None,
    ) -> torch.Tensor:
        """
        Computes the log-probabilities of the input data.

        Args:
            x: input data, shape (n, dims_in)
            c: condition, shape (n, dims_c) or None for an unconditional flow
            channel: encodes the channel of the samples. It must have one of the following types:

                - ``Tensor``: integer tensor of shape (n, ), containing the channel index for every
                  input sample;
                - ``list``: list of integers, specifying the number of samples in each channel;
                - ``int``: integer specifying a single channel containing all the samples;
                - ``None``: used in the single-channel case or to indicate that all channels contain
                  the same number of samples in the multi-channel case.
        Returns:
            log-probabilities with shape (n, )
        """
        if self.discrete_dims_first:
            x_discrete = x[:, : self.dims_in_discrete].long()
            log_prob_discrete = self.discrete_flow.log_prob(
                x_discrete, c=c, channel=channel
            )
            condition = self.encode_discrete(x_discrete)
            if c is not None:
                condition = torch.cat((c, condition), dim=1)
            log_prob_continuous = self.continuous_flow.log_prob(
                x[:, self.dims_in_discrete :], c=condition, channel=channel
            )
        else:
            x_continuous = x[:, : self.dims_in_continuous]
            log_prob_continuous = self.continuous_flow.log_prob(
                x_continuous, c=c, channel=channel
            )
            condition = (
                x_continuous if c is None else torch.cat((c, x_continuous), dim=1)
            )
            log_prob_discrete = self.discrete_flow.log_prob(
                x[:, self.dim_in_continuous].long(), c=condition, channel=channel
            )
        return log_prob_discrete + log_prob_continuous

    def sample(
        self,
        n: int | None = None,
        c: torch.Tensor | None = None,
        channel: torch.Tensor | list[int] | int | None = None,
        return_log_prob: bool = False,
        return_prob: bool = False,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ) -> torch.Tensor | tuple[torch.Tensor, ...]:
        """ """
        if self.discrete_dims_first:
            x_discrete, prob_discrete = self.discrete_flow.sample(
                n=n,
                c=c,
                channel=channel,
                return_prob=True,
                device=device,
                dtype=dtype,
            )
            condition = self.encode_discrete(x_discrete)
            if c is not None:
                condition = torch.cat((c, condition), dim=1)
            x_continuous, log_prob_continuous = self.continuous_flow.sample(
                c=condition,
                channel=channel,
                return_log_prob=True,
                device=device,
                dtype=dtype,
            )
            x = torch.cat((x_discrete.to(x_continuous.dtype), x_continuous), dim=1)
        else:
            x_continuous, log_prob_continuous = self.continuous_flow.sample(
                n=n,
                c=c,
                channel=channel,
                return_log_prob=True,
                device=device,
                dtype=dtype,
            )
            condition = (
                x_continuous if c is None else torch.cat((c, x_continuous), dim=1)
            )
            x_discrete, prob_discrete = self.discrete_flow.sample(
                c=condition,
                channel=channel,
                return_prob=True,
                device=device,
                dtype=dtype,
            )
            x = torch.cat((x_continuous.to(x_continuous.dtype), x_discrete), dim=1)

        extra_returns = []
        if return_log_prob:
            extra_returns.append(prob_discrete.log() + log_prob_continuous)
        if return_prob:
            extra_returns.append(prob_discrete * log_prob_continuous.exp())
        if len(extra_returns) > 0:
            return x, *extra_returns
        else:
            return x



from collections.abc import Callable

import torch
import torch.nn as nn
import torch.nn.functional as F

from .flow import Distribution
from .mlp import MLP


class DiscreteTransformer(nn.Module, Distribution):
    def __init__(
        self,
        dims_in: list[int],
        dims_c: int = 0,
        prior_prob_function: Callable[[torch.Tensor, int], torch.Tensor] | None = None,
        embedding_dim: int = 64,
        feedforward_dim: int = 256,
        heads: int = 8,
        transformer_layers: int = 3,
        mlp_layers: int = 3,
        mlp_units: int = 256,
        mlp_activation: Callable[[], nn.Module] = nn.ReLU,
    ):
        super().__init__()
        self.dims_in = dims_in
        self.prior_prob_function = prior_prob_function

        self.max_dim = max(dims_in)
        n_dims = len(dims_in)
        self.transformer = nn.TransformerEncoder(
            encoder_layer=nn.TransformerEncoderLayer(
                d_model=embedding_dim,
                nhead=heads,
                dim_feedforward=feedforward_dim,
                dropout=0,
                activation="gelu",
                batch_first=True,
            ),
            num_layers=transformer_layers,
        )
        self.mlp = MLP(
            features_in=embedding_dim,
            features_out=self.max_dim,
            layers=mlp_layers,
            units=mlp_units,
            activation=mlp_activation,
        )
        self.x_embedding = nn.Embedding(
            num_embeddings=self.max_dim, embedding_dim=embedding_dim
        )
        self.c_embedding = nn.Linear(dims_c, embedding_dim) if dims_c > 0 else None
        self.pos_embedding = nn.Embedding(
            num_embeddings=n_dims, embedding_dim=embedding_dim
        )

        self.register_buffer(
            "causal_mask",
            torch.ones((n_dims, n_dims), dtype=torch.bool).triu(diagonal=1),
        )
        prior_mask = torch.zeros((n_dims, self.max_dim))
        for i, n_opts in enumerate(self.dims_in):
            prior_mask[i, :n_opts] = 1
        self.register_buffer("prior_mask", prior_mask)

    def log_prob(
        self,
        x: torch.Tensor,
        c: torch.Tensor | None = None,
        channel: torch.Tensor | list[int] | int | None = None,
    ) -> torch.Tensor:
        x_embed = F.pad(self.x_embedding(x[:, :-1]), (0, 0, 1, 0))
        pos_embed = self.pos_embedding(
            torch.arange(len(self.dims_in), device=x.device)[None, :]
        )
        embedding = x_embed + pos_embed
        if c is not None:
            embedding += self.c_embedding(c)
        net_log_probs = self.mlp(self.transformer(embedding, self.causal_mask))

        prior_probs = self.prior_mask.expand(x.shape[0], -1, -1).clone()
        if self.prior_prob_function is not None:
            for i, n_opts in enumerate(self.dims_in):
                prior_probs[:, i, :n_opts] = self.prior_prob_function(x[:, :i], i)

        probs_unnorm = net_log_probs.exp() * prior_probs
        prob = torch.prod(
            torch.gather(probs_unnorm, 2, x[:, :, None])[:, :, 0]
            / probs_unnorm.sum(dim=2),
            dim=1,
        )

        return prob.log()

    def sample(
        self,
        n: int | None = None,
        c: torch.Tensor | None = None,
        channel: torch.Tensor | list[int] | int | None = None,
        return_log_prob: bool = False,
        return_prob: bool = False,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ) -> torch.Tensor | tuple[torch.Tensor, ...]:
        if n is None:
            n = len(c)
        if c is not None:
            dtype = c.dtype
            device = c.device

        x = torch.zeros((n, 0), dtype=torch.int64, device=device)
        prob = torch.ones((n,), dtype=dtype, device=device)
        embedding = None
        if c is not None:
            c_embedding = self.c_embedding(c)
        for i, dim in enumerate(self.dims_in):
            next_embedding = (
                self.pos_embedding(torch.tensor([i], device=device))
                .expand(n, -1)
                .clone()
            )
            if i > 0:
                next_embedding += self.x_embedding(x[:, i - 1])
            if c is not None:
                next_embedding += c_embedding
            if embedding is None:
                embedding = next_embedding[:, None, :]
            else:
                embedding = torch.cat((embedding, next_embedding[:, None, :]), dim=1)
            net_log_probs = self.mlp(
                self.transformer(embedding, self.causal_mask[: i + 1, : i + 1])[:, -1]
            )

            unnorm_probs = net_log_probs.exp() * self.prior_mask[None, i, :]
            if self.prior_prob_function is not None:
                unnorm_probs[:, :dim] *= self.prior_prob_function(x[:, :i], i)
            cdf = unnorm_probs.cumsum(dim=1)
            norm = cdf[:, -1]
            cdf = cdf / norm[:, None]
            r = torch.rand((n, 1), dtype=dtype, device=device)
            samples = torch.searchsorted(cdf, r)
            x = torch.cat((x, samples), dim=1)
            prob = prob * torch.gather(unnorm_probs, 1, samples)[:, 0] / norm

        return_list = [x]
        if return_log_prob:
            return_list.append(prob.log())
        if return_prob:
            return_list.append(prob)
        if len(return_list) > 1:
            return return_list
        else:
            return return_list[0]



import itertools
import math

import torch
import torch.nn as nn
import torch.nn.functional as F


class MaskedMLP(nn.Module):
    def __init__(
        self,
        input_dims: list[int],
        output_dims: list[int],
        layers: int = 3,
        nodes_per_feature: int = 8,
        activation=nn.LeakyReLU,
        channels: int | None = None,
    ):
        super().__init__()

        in_degrees = []
        for i, in_dim in enumerate(input_dims):
            in_degrees.extend([i] * in_dim)
        hidden_degrees = torch.repeat_interleave(
            torch.arange(len(input_dims)), nodes_per_feature
        )
        out_degrees = []
        for i, out_dim in enumerate(output_dims):
            out_degrees.extend([i] * out_dim)
        hidden_layers = layers - 1
        layer_degrees = [
            torch.tensor(in_degrees),
            *([hidden_degrees] * hidden_layers),
            torch.tensor(out_degrees),
        ]

        self.in_slices = [[slice(0)] * layers]
        self.out_slices = [[slice(0)] * layers]
        hidden_dims = [nodes_per_feature] * hidden_layers
        for in_dim, out_dim in zip(input_dims, output_dims):
            self.in_slices.append(
                [
                    slice(0, prev_slice_in.stop + deg_in)
                    for deg_in, prev_slice_in in zip(
                        [in_dim, *hidden_dims], self.in_slices[-1]
                    )
                ]
            )
            self.out_slices.append(
                [
                    slice(prev_slice_out.stop, prev_slice_out.stop + deg_out)
                    for deg_out, prev_slice_out in zip(
                        [*hidden_dims, out_dim], self.out_slices[-1]
                    )
                ]
            )
        self.in_slices.pop(0)
        self.out_slices.pop(0)

        self.channels = 1 if channels is None else channels
        self.masks = nn.ParameterList()
        self.weights = nn.ModuleList()
        self.biases = nn.ModuleList()

        for i in range(self.channels):
            chan_weights = nn.ParameterList()
            chan_biases = nn.ParameterList()
            for deg_in, deg_out in zip(layer_degrees[:-1], layer_degrees[1:]):
                if i == 0:
                    self.masks.append(
                        nn.Parameter(
                            (deg_out[:, None] >= deg_in[None, :]).float(),
                            requires_grad=False,
                        )
                    )
                chan_weights.append(
                    nn.Parameter(torch.empty((len(deg_out), len(deg_in))))
                )
                chan_biases.append(nn.Parameter(torch.empty((len(deg_out),))))
            self.weights.append(chan_weights)
            self.biases.append(chan_biases)

        self.activation = activation()
        self.reset_parameters()

    def reset_parameters(self):
        for chan_weights, chan_biases in zip(self.weights, self.biases):
            for weight, bias in zip(chan_weights[:-1], chan_biases[:-1]):
                torch.nn.init.kaiming_uniform_(weight, a=math.sqrt(5))
                fan_in, _ = torch.nn.init._calculate_fan_in_and_fan_out(weight)
                bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
                torch.nn.init.uniform_(bias, -bound, bound)
            nn.init.zeros_(chan_weights[-1])
            nn.init.zeros_(chan_biases[-1])

    def forward(
        self,
        x: torch.Tensor,
        channel: list[int] | int | None = None,
    ) -> torch.Tensor:
        if self.channels == 1:
            return self._forward_single(x, 0)
        elif isinstance(channel, list):
            return torch.cat(
                [
                    self._forward_single(xi, i)
                    for i, xi in enumerate(x.split(channel, dim=0))
                ],
                dim=0,
            )
        elif isinstance(channel, int):
            return self._forward_single(x, channel)
        else:
            # TODO: implement optimized version for the uniform special case
            return torch.cat(
                [
                    self._forward_single(xi, i)
                    for i, xi in enumerate(x.split(self.channels, dim=0))
                ],
                dim=0,
            )

    def forward_cached(
        self,
        x: torch.Tensor,
        feature: int,
        caches: list[list[torch.Tensor] | None] | None = None,
        channel: list[int] | int | None = None,
    ) -> tuple[torch.Tensor, list[list[torch.Tensor] | None]]:
        caches = itertools.repeat(None) if caches is None else caches
        if self.channels == 1:
            x, cache = self._forward_cached_single(x, feature, next(iter(caches)), 0)
            return x, [cache]
        elif isinstance(channel, list):
            xs, caches = zip(
                *[
                    self._forward_cached_single(xi, feature, cache, i)
                    for i, (xi, cache) in enumerate(
                        zip(x.split(channel, dim=0), caches)
                    )
                ]
            )
            return torch.cat(xs, dim=0), caches
        elif isinstance(channel, int):
            x, cache = self._forward_cached_single(
                x, feature, next(iter(caches)), channel
            )
            return x, [cache]
        else:
            # TODO: implement optimized version for the uniform special case
            xs, caches = zip(
                *[
                    self._forward_cached_single(xi, feature, cache, i)
                    for i, (xi, cache) in enumerate(
                        zip(x.split(self.channels, dim=0), caches)
                    )
                ]
            )
            return torch.cat(xs, dim=0), caches

    def _forward_single(self, x: torch.Tensor, channel: int) -> torch.Tensor:
        weights, biases = self.weights[channel], self.biases[channel]
        for weight, bias, mask in zip(weights[:-1], biases[:-1], self.masks[:-1]):
            x = self.activation(F.linear(x, mask * weight, bias))
        return F.linear(x, self.masks[-1] * weights[-1], biases[-1])

    def _forward_cached_single(
        self,
        x: torch.Tensor,
        feature: int,
        cache: list[torch.Tensor] | None,
        channel: int,
    ) -> tuple[torch.Tensor, list[torch.Tensor]]:
        weights, biases = self.weights[channel], self.biases[channel]
        if cache is None:
            cache = [None] * len(weights)
        new_cache = []
        in_slices = self.in_slices[feature]
        out_slices = self.out_slices[feature]
        first = True
        for weight, bias, in_slice, out_slice, x_cached in zip(
            weights, biases, in_slices, out_slices, cache
        ):
            if first:
                first = False
            else:
                x = self.activation(x)
            if x_cached is not None:
                x = torch.cat((x_cached, x), dim=1)
            new_cache.append(x)
            x = F.linear(x, weight[out_slice, in_slice], bias[out_slice])
        return x, new_cache
